from typing import Optional, Union, Tuple
from functools import partial
from enum import Enum
import pandas as pd
import requests
import pyterrier as pt
import pyterrier_alpha as pta
from . import http_error_retry, paginated_search, multi_query


class DblpEntityType(Enum):
    publication = 'publication'
    author = 'author'
    venue = 'venue'


class DblpBibType(Enum):
    standard = 'standard'
    condensed = 'condensed'
    with_crossref = 'with_crossref'


class DblpApi:
    """Represents a reference to the DBLP search API."""

    API_BASE_URL = 'https://dblp.org'

    def retriever(self,
        *,
        num_results: int = 100,
        entity_type: Union[str, DblpEntityType] = DblpEntityType.publication,
        verbose: bool = True
    ) -> pt.Transformer:
        """Returns a :class:`~pyterrier.Transformer` that retrieves from DBLP.

        Args:
            num_results: The number of results to retrieve. Defaults to 100.
            entity_type: The type of entity to search over. Defaults to ``DblpEntityType.publication``.
            verbose: Whether to log the progress. Defaults to True.
        """
        return DblpRetriever(api=self, num_results=num_results, entity_type=entity_type, verbose=verbose)

    def bibtex_loader(self,
        *,
        bib_type: Union[str, DblpBibType] = DblpBibType.standard,
        verbose: bool = True
    ) -> pt.Transformer:
        """Returns a :class:`~pyterrier.Transformer` that loads bibtex data from DBLP.

        Args:
            bib_type: The type of BibTeX to load. Defaults to ``DblpBibType.standard``.
            verbose: Whether to log the progress. Defaults to True.
        """
        return DblpBibtexLoader(api=self, bib_type=bib_type, verbose=verbose)

    def search(self,
        query: str,
        *,
        entity_type: Union[str, DblpEntityType] = DblpEntityType.publication,
        offset: int = 0,
        limit: int = 100,
        return_next: bool = False,
        return_total: bool = False,
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, int], Tuple[pd.DataFrame, int, int]]:
        """Searches for papers on Semantic Scholar with the provided query.

        Args:
            query: The search query.
            entity_type: The type of entity to search over. Defaults to ``DblpEntityType.publication``.
            offset: The offset of the first result to retrieve. Defaults to 0.
            limit: The maximum number of results to retrieve. Defaults to 100.
            return_next: Whether to return the next query URL. Defaults to False.
            return_total: Whether to return the total number of results. Defaults to False.
        """
        entity_type = DblpEntityType(entity_type)
        limit = max(min(limit, 1000), 1)
        params = {
            'q': query,
            'format': 'json',
            'f': offset,
            'c': limit,
        }
        endpoint = {
            DblpEntityType.publication: '/search/publ/api',
            DblpEntityType.author: '/search/author/api',
            DblpEntityType.venue: '/search/venue/api',
        }[entity_type]
        http_res = requests.get(DblpApi.API_BASE_URL + endpoint, params=params)
        http_res.raise_for_status()
        http_res = http_res.json()['result']

        data_columns, docno_column_fn = {
            DblpEntityType.publication: (['title', 'authors', 'year', 'type'], lambda hit: hit['info']['key']),
            DblpEntityType.author: (['author'], lambda hit: hit['info']['url'].replace('https://dblp.org/pid/', '')),
            DblpEntityType.venue: (['venue', 'acronym', 'type'], lambda hit: hit['info']['url'].replace('https://dblp.org/db/', '')[:-1]),
        }[entity_type]

        result_df = []
        first = int(http_res['hits']['@first'])
        for rank, hit in zip(range(limit), http_res['hits'].get('hit', [])):
            row = [
                docno_column_fn(hit),
                first + rank,
                -1.0 * (first + rank),
            ]
            for c in data_columns:
                if c == 'authors':
                    authors = hit['info'][c].get('author', [])
                    if isinstance(authors, dict):
                        authors = [authors]
                    row.append([a['text'] for a in authors])
                elif c == 'title':
                    row.append(hit['info'][c].rstrip('.')) # dblp search results add a trailing . to titles for some reason?
                else:
                    row.append(hit['info'].get(c))
            result_df.append(row)
        result_df = pd.DataFrame(result_df, columns=['docno', 'rank', 'score', *data_columns])

        res = [result_df]
        if return_next:
            res.append(first + int(http_res['hits']['@sent']))
        if return_total:
            res.append(int(http_res['hits']['@total']))
        if len(res) == 1:
            return res[0]
        return tuple(res)

    def load_bibtex(self,
        docno: str,
        *,
        bib_type: Union[str, DblpBibType] = DblpBibType.standard,
    ) -> str:
        bib_type = DblpBibType(bib_type)
        param = {
            DblpBibType.standard: '1',
            DblpBibType.condensed: '0',
            DblpBibType.with_crossref: '2',
        }[bib_type]
        http_res = requests.get(f'{DblpApi.API_BASE_URL}/rec/{docno}.bib', params={'param': param})
        http_res.raise_for_status()
        return http_res.text


class DblpRetriever(pt.Transformer):
    """A :class:`~pyterrier.Transformer` retriever that queries the DBLP search API."""
    def __init__(self,
        *,
        api: Optional[DblpApi] = None,
        num_results: int = 100,
        entity_type: Union[str, DblpEntityType] = DblpEntityType.publication,
        verbose: bool = True
    ):
        """
        Args:
            api: The DBLP api service. Defaults to a new instance of :class:`~pyterrier_services.DblpApi`.
            num_results: The number of results to retrieve per query. Defaults to 100.
            entity_type: The type of entity to search over. Defaults to ``DblpEntityType.publication``
            verbose: Whether to log the progress. Defaults to True.
        """
        self.api = api or DblpApi()
        self.num_results = num_results
        self.entity_type = entity_type
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.query_frame(inp, extra_columns=['query'])
        return multi_query(
            paginated_search(
                http_error_retry(
                    partial(self.api.search, entity_type=self.entity_type)
                ),
                num_results=self.num_results,
            ),
            verbose=self.verbose,
            verbose_desc='DblpRetriever',
        )(inp)

    def fuse_rank_cutoff(self, k: int) -> Optional['DblpRetriever']:
        if k < self.num_results:
            return DblpRetriever(api=self.api, num_results=k, entity_type=self.entity_type, verbose=self.verbose)


class DblpBibtexLoader(pt.Transformer):
    """A :class:`~pyterrier.Transformer` that loads BibTeX data from DBLP."""
    def __init__(self,
        *,
        api: Optional[DblpApi] = None,
        bib_type: Union[str, DblpBibType] = DblpBibType.standard,
        verbose: bool = True
    ):
        """
        Args:
            api: The DBLP api service. Defaults to a new instance of :class:`~pyterrier_services.DblpApi`.
            bib_type: The type of BibTeX to load. Defaults to ``DblpBibType.standard``.
            verbose: Whether to log the progress. Defaults to True.
        """
        self.api = api or DblpApi()
        self.bib_type = bib_type
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.columns(inp, includes=['docno'])
        bibtex = []
        it = inp['docno']
        if self.verbose:
            it = pt.tqdm(it, desc='DblpBibtexLoader')
        for docno in it:
            bibtex.append(self.api.load_bibtex(docno, bib_type=self.bib_type))
        return inp.assign(bibtex=bibtex)



import unittest
import pandas as pd
from pyterrier_services import GoogleApi

class TestGoogle(unittest.TestCase):
    def test_retriever(self):
        google = GoogleApi()
        retr = google.retriever(num_results=15)
        res = retr.search('PyTerrier')
        self.assertIsInstance(res, pd.DataFrame)
        self.assertEqual(len(res), 15)
        self.assertEqual(set(res.columns), {'qid', 'query', 'docno', 'score', 'rank', 'url', 'title', 'snippet'})



__version__ = '0.4.2'

from .core import http_error_retry, paginated_search, multi_query
from .semantic_scholar import SemanticScholarApi, SemanticScholarRetriever
from .pinecone import PineconeApi, PineconeSparseModel, PineconeDenseModel, PineconeReranker
from .dblp import DblpApi, DblpRetriever
from .google import GoogleApi, GoogleSearchRetriever

__all__ = [
	'http_error_retry', 'paginated_search', 'multi_query',
	'SemanticScholarApi', 'SemanticScholarRetriever',
	'PineconeApi', 'PineconeSparseModel', 'PineconeDenseModel', 'PineconeReranker',
	'DblpApi', 'DblpRetriever',
	'GoogleApi', 'GoogleSearchRetriever',
]



from typing import Optional, Union, Tuple
import os
import pandas as pd
import pyterrier as pt
from pyterrier_services import paginated_search, multi_query

_HELP_URL = 'https://developers.google.com/custom-search/v1/overview'

class GoogleApi:
    """Represents a refernece to the Google API."""

    def __init__(self, api_key: Optional[str] = None):
        """
        Args: 
            api_key (str): the Google API key (taken from ``GOOGLE_API_KEY`` env variable if not provided)
        """
        if api_key is None:
            api_key = os.environ.get("GOOGLE_API_KEY")
        if api_key is None:
            raise ValueError(f"A Google API key must be specified (either as GOOGLE_API_KEY env variable or passed to `GoogleApi(api_key='...')`). See <{_HELP_URL}> for details on how to get an API key.")
        self.api_key = api_key

        try:
            from googleapiclient.discovery import build
        except ModuleNotFoundError as mnfe:
            raise Exception("You need to pip install google-api-python-client") from mnfe
        self._build = build

    def retriever(self, cx: Optional[str] = None, *, num_results: int = 10, verbose: bool = False) -> pt.Transformer:
        """Creates a :class:`GoogleSearchRetriever` instance, allowing retrieval over the Google search engine.

        Follow Google's guide for a `Custom Search JSON API <{_HELP_URL}>`_ to get
        the right parameters.

        Arguments:
            cx (str): the service to access (taken from ``GOOGLE_CSE_CX`` env variable if not provided)
            num_results (int): The number of results to retrieve per query. Defaults to 10.

        Returns:
            :class:`pyterrier.Transformer`: A PyTerrier transformer that can be used to
            perform Google searches.

        Output columns: ['title', 'url', 'snippet']

        Example::
            title               Chemical reaction | Definition, Equations, Exa...
            url                 https://www.britannica.com/science/chemical-re...
            snippet             Mar 24, 2025 ... A chemical reaction is a proc...
        """.format(_HELP_URL=_HELP_URL)
        return GoogleSearchRetriever(self, cx, num_results=num_results, verbose=verbose)


class GoogleSearchRetriever(pt.Transformer):
    """A :class:`~pyterrier.Transformer` retriever that queries the Google search API."""

    def __init__(self,
        api: Optional[GoogleApi] = None,
        cx: Optional[str] = None,
        *,
        num_results: int = 10,
        verbose: bool = False,
    ):
        """
        Args:
            api (GoogleApi): The Google API service. Defaults to a new instance of :class:`GoogleApi`.
            cx: (str): The Google Custom Search Engine ID. This is required to perform searches.
            num_results (int): The number of results to retrieve per query. Defaults to 10.
            verbose (bool): Whether to log the progress. Defaults to False.
        """
        self.api = api or GoogleApi()
        if cx is None:
            cx = os.getenv("GOOGLE_CSE_CX")
        if cx is None:
            raise ValueError(f"A Google Custom Search Engine ID (cx) must be specified. See <{_HELP_URL}> for details on how to get a Custom Search Engine ID.")
        self.cx = cx
        self.cse_service = self.api._build("customsearch", "v1", developerKey=self.api.api_key).cse()
        self.num_results = num_results
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        return multi_query(
            paginated_search(self._search_internal, num_results=self.num_results),
            verbose=self.verbose,
            verbose_desc='GoogleSearchRetriever',
        )(inp)

    def _search_internal(self,
        query: str,
        *,
        offset: int = 0,
        limit: int = 10,
        return_next: bool = False,
        return_total: bool = False
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, int], Tuple[pd.DataFrame, int, int]]:
        """Searches for papers on Google with the provided query.

        Args:
            query: The search query.
            offset: The offset of the first result to retrieve. Defaults to 0.
            limit: The maximum number of results to retrieve. Defaults to 10.
            return_next: Whether to return the next query URL. Defaults to False.
            return_total: Whether to return the total number of results. Defaults to False.
        """
        api_result = self.cse_service.list(q=query, cx=self.cx, num=min(limit, 10), start=offset).execute()
        if len(api_result["items"]) == 0:
            result_df = pd.DataFrame(columns=['docno', 'url', 'title', 'snippet', 'rank', 'score'])
        else:
            result_df = pd.DataFrame([[r['link'], r['link'], r['title'], r['snippet']] for r in api_result["items"]], columns=['docno', 'url', 'title', 'snippet'])
            result_df['rank'] = range(offset, offset+len(result_df))
            result_df['score'] = -result_df['rank']

        res = [result_df]
        if return_next:
            res.append(offset + len(result_df) + 1)
        if return_total:
            res.append(int(res['searchInformation']['totalResults']))
        if len(res) == 1:
            return res[0]
        return tuple(res)

    def fuse_rank_cutoff(self, k: int) -> Optional['GoogleSearchRetriever']:
        if k < self.num_results:
            return GoogleSearchRetriever(api=self.api, cx=self.cx, num_results=k, verbose=self.verbose)



from typing import Optional, Literal
import numpy as np
import pandas as pd
import pyterrier as pt
import pyterrier_alpha as pta

class PineconeApi:
    """Represents a reference to the Pinecone API.

    This class wraps :class:`pinecone.Pinecone`.
    """

    def __init__(self, api_key: Optional[str] = None):
        """
        Args:
            api_key (str, optional): The Pinecone API key. Defaults to the value from ``PINECONE_API_KEY``.
        """
        from pinecone import Pinecone
        self.api_key = api_key
        self.pc = Pinecone(api_key=self.api_key)
        self._embed = self.pc.inference.embed
        self._rerank = self.pc.inference.rerank

    def dense_model(self, model_name: str = 'multilingual-e5-large') -> 'PineconeDenseModel':
        """Creates a :class:`PineconeDenseModel` instance.

        Args:
            model_name (str): The name of the model. See the `list of supported models <https://docs.pinecone.io/models>`__.
        """
        return PineconeDenseModel(model_name, api=self)

    def sparse_model(self, model_name: str = 'pinecone-sparse-english-v0') -> 'PineconeSparseModel':
        """Creates a :class:`PineconeSparseModel` instance.

        Args:
            model_name (str): The name of the model. See the `list of supported models <https://docs.pinecone.io/models>`__.
        """
        return PineconeSparseModel(model_name, api=self)

    def reranker(self, model_name: str = 'pinecone-rerank-v0') -> 'PineconeReranker':
        """Creates a :class:`PineconeReranker` instance.

        Args:
            model_name (str): The name of the model. See the `list of supported models <https://docs.pinecone.io/models>`__.
        """
        return PineconeReranker(model_name, api=self)


class PineconeSparseModel(pt.Transformer):
    """A PyTerrier transformer that provies access to a Pinecone sparse model."""
    def __init__(self,
        model_name: str = 'pinecone-sparse-english-v0',
        *,
        api: Optional[PineconeApi] = None,
    ):
        """
        Args:
            model_name (str): The name of the model. See the `list of supported models <https://docs.pinecone.io/models>`__.
            api (PineconeApi, optional): The Pinecone API object. Defaults to a new instance.
        """
        self.model_name = model_name
        self.api = api or PineconeApi()

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """Encodes either queries or documents using this model (based on input columns)"""
        with pta.validate.any(inp) as v:
            v.query_frame(extra_columns=['query'], mode=self.query_encoder)
            v.document_frame(extra_columns=['text'], mode=self.doc_encoder)
            v.result_frame(extra_columns=['query', 'text'], mode=self.scorer)
        return v.mode()(inp)

    def query_encoder(self) -> 'PineconeSparseEncoder':
        """Creates a transformer that encodes queries using this model."""
        return PineconeSparseEncoder(self, input_type='query')

    def doc_encoder(self) -> 'PineconeSparseEncoder':
        """Creates a transformer that encodes documents using this model."""
        return PineconeSparseEncoder(self, input_type='passage')

    def scorer(self) -> 'PineconeSparseScorer':
        """Creates a transformer that scores (re-ranks) results using this model."""
        return PineconeSparseScorer(self)

    def __repr__(self):
        return f"PineconeSparseModel({self.model_name!r})"


class PineconeSparseEncoder(pt.Transformer):
    def __init__(self, sparse_model: PineconeSparseModel, *, input_type: Literal['passage', 'query']):
        self.sparse_model = sparse_model
        self.input_type = input_type

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        if self.input_type == 'passage':
            pta.validate.document_frame(inp, extra_columns=['text'])
            text = inp['text'].tolist()
            toks_field = 'toks'
        elif self.input_type == 'query':
            pta.validate.query_frame(inp, extra_columns=['query'])
            text = inp['query'].tolist()
            toks_field = 'query_toks'

        embeddings = self.sparse_model.api._embed(
            model=self.sparse_model.model_name,
            inputs=text,
            parameters={"input_type": self.input_type, "return_tokens": True}
        )
        assert embeddings.vector_type == 'sparse'
        toks = [dict(zip(v.sparse_tokens, v.sparse_values)) for v in embeddings.data]
        return inp.assign(**{toks_field: toks})

    def __repr__(self):
        return f"PineconeSparseEncoder({self.sparse_model!r}, input_type={self.input_type!r})"


def _sparse_dot(qt, dt):
    return sum(qt[t] * dt[t] for t in qt.keys() & dt.keys())


class PineconeSparseScorer(pt.Transformer):
    def __init__(self, sparse_model: PineconeSparseModel):
        self.sparse_model = sparse_model

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.document_frame(inp, extra_columns=['query', 'text'])
        query_toks = self.sparse_model.query_encoder(inp[['qid', 'query']])['query_toks']
        doc_toks = self.sparse_model.doc_encoder(inp[['docno', 'text']])['toks']

        scores = [_sparse_dot(qt, dt) for qt, dt in zip(query_toks, doc_toks)]

        res = inp.assign(score=scores).sort_values('score', ascending=False).reset_index(drop=True)
        pt.model.add_ranks(res)
        return res

    def __repr__(self):
        return f"PineconeSparseScorer({self.sparse_model!r})"


class PineconeReranker(pt.Transformer):
    """A PyTerrier transformer that provies access to a Pinecone reranker model."""

    def __init__(self,
        model_name: str = 'pinecone-rerank-v0',
        *,
        api: Optional[PineconeApi] = None,
    ):
        """
        Args:
            model_name (str): The name of the model. See the `list of supported models <https://docs.pinecone.io/models>`__.
            api (PineconeApi, optional): The Pinecone API object. Defaults to a new instance.
        """
        self.model_name = model_name
        self.api = api or PineconeApi()

    @pta.transform.by_query()
    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        inp = inp.reset_index(drop=True)
        pta.validate.result_frame(inp, extra_columns=['query', 'text'])
        query = inp['query'].iloc[0]
        documents = inp['text'].tolist()
        results = self.api._rerank(
            model=self.model_name,
            query=query,
            documents=documents,
            return_documents=False,
            parameters= {
                "truncate": "END"
            }
        )
        res = inp.assign(score=pd.Series(
            [r.score for r in results.data],
            [r.index for r in results.data],
        ))
        res = res.sort_values('score', ascending=False).reset_index(drop=True)
        res = res.assign(rank=np.arange(len(res)))
        return res

    def __repr__(self):
        return f"PineconeReranker({self.model_name!r})"


class PineconeDenseModel(pt.Transformer):
    """A PyTerrier transformer that provides access to a Pinecone dense model."""
    def __init__(self,
        model_name: str = 'multilingual-e5-large',
        *,
        api: Optional[PineconeApi] = None,
    ):
        """
        Args:
            model_name (str): The name of the model. See the `list of supported models <https://docs.pinecone.io/models>`__.
            api (PineconeApi, optional): The Pinecone API object. Defaults to a new instance.
        """
        self.model_name = model_name
        self.api = api or PineconeApi()

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """Encodes either queries or documents using this model (based on input columns)"""
        with pta.validate.any(inp) as v:
            v.query_frame(extra_columns=['query'], mode=self.query_encoder)
            v.document_frame(extra_columns=['text'], mode=self.doc_encoder)
            v.result_frame(extra_columns=['query', 'text'], mode=self.scorer)
        return v.mode()(inp)

    def query_encoder(self) -> 'PineconeDenseEncoder':
        """Creates a transformer that encodes queries using this model."""
        return PineconeDenseEncoder(self, input_type='query')

    def doc_encoder(self) -> 'PineconeDenseEncoder':
        """Creates a transformer that encodes documents using this model."""
        return PineconeDenseEncoder(self, input_type='passage')

    def scorer(self) -> 'PineconeDenseScorer':
        """Creates a transformer that scores (re-ranks) results using this model."""
        return PineconeDenseScorer(self)

    def __repr__(self):
        return f"PineconeDenseModel({self.model_name!r})"


class PineconeDenseEncoder(pt.Transformer):
    def __init__(self, dense_model: PineconeDenseModel, *, input_type: Literal['passage', 'query']):
        self.dense_model = dense_model
        self.input_type = input_type

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        if self.input_type == 'passage':
            pta.validate.document_frame(inp, extra_columns=['text'])
            text = inp['text'].tolist()
            vecs_field = 'doc_vec'
        elif self.input_type == 'query':
            pta.validate.query_frame(inp, extra_columns=['query'])
            text = inp['query'].tolist()
            vecs_field = 'query_vec'

        embeddings = self.dense_model.api._embed(
            model=self.dense_model.model_name,
            inputs=text,
            parameters={"input_type": self.input_type, "truncate": "END"}
        )
        assert embeddings.vector_type == 'dense'
        vecs = [np.array(v.values) for v in embeddings.data]
        return inp.assign(**{vecs_field: vecs})

    def __repr__(self):
        return f"PineconeDenseEncoder({self.dense_model!r}, input_type={self.input_type!r})"


class PineconeDenseScorer(pt.Transformer):
    def __init__(self, dense_model: PineconeDenseModel):
        self.dense_model = dense_model

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        pta.validate.document_frame(inp, extra_columns=['query', 'text'])
        query_vecs = self.dense_model.query_encoder(inp[['qid', 'query']])['query_vec']
        doc_vecs = self.dense_model.doc_encoder(inp[['docno', 'text']])['doc_vec']

        scores = [np.dot(qv, dv) for qv, dv in zip(query_vecs, doc_vecs)]

        res = inp.assign(score=scores).sort_values('score', ascending=False).reset_index(drop=True)
        pt.model.add_ranks(res)
        return res

    def __repr__(self):
        return f"PineconeDenseScorer({self.dense_model!r})"



from typing import List, Optional, Union, Tuple
from functools import partial
import pandas as pd
import requests
import pyterrier as pt
from . import http_error_retry, paginated_search, multi_query

class SemanticScholarApi:
    """Represents a reference to the Semantic Scholar search API."""
    API_BASE_URL = 'https://api.semanticscholar.org/graph/v1'

    def retriever(self,
        *,
        num_results: int = 100,
        fields: List[str] = ['title', 'abstract'],
        verbose: bool = True
    ) -> pt.Transformer:
        """Returns a :class:`~pyterrier.Transformer` that retrieves articles from Semantic Scholar.

        Args:
            num_results: The number of results to retrieve. Defaults to 100.
            fields: The fields to include in the retrieved results. Defaults to ['title', 'abstract'].
            verbose: Whether to log the progress. Defaults to True.
        """
        return SemanticScholarRetriever(api=self, num_results=num_results, fields=fields, verbose=verbose)

    def search(self,
        query: str,
        *,
        offset: int = 0,
        limit: int = 100,
        fields: List[str] = ['title', 'abstract'],
        return_next: bool = False,
        return_total: bool = False
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, int], Tuple[pd.DataFrame, int, int]]:
        """Searches for papers on Semantic Scholar with the provided query.

        Args:
            query: The search query.
            offset: The offset of the first result to retrieve. Defaults to 0.
            limit: The maximum number of results to retrieve. Defaults to 100.
            fields: The fields to include in the retrieved results. Defaults to ['title', 'abstract'].
            return_next: Whether to return the next query URL. Defaults to False.
            return_total: Whether to return the total number of results. Defaults to False.
        """
        params = {
            'query': query,
            'offset': offset,
            'fields': ','.join(fields),
            'limit': max(min(limit, 100), 1),
        }
        http_res = requests.get(SemanticScholarApi.API_BASE_URL + '/paper/search', params=params)
        http_res.raise_for_status()
        http_res = http_res.json()

        if len(http_res['data']) == 0:
            result_df = pd.DataFrame(columns=['docno', *[str(f) for f in fields], 'rank', 'score'])
        else:
            result_df = pd.DataFrame(http_res['data'])
            result_df.rename(columns={'paperId': 'docno'}, inplace=True)
            result_df['rank'] = range(http_res['offset'], http_res['offset']+len(result_df))
            result_df['score'] = -result_df['rank']

        res = [result_df]
        if return_next:
            res.append(http_res.get('next'))
        if return_total:
            res.append(http_res['total'])
        if len(res) == 1:
            return res[0]
        return tuple(res)


class SemanticScholarRetriever(pt.Transformer):
    """A :class:`~pyterrier.Transformer` retriever that queries the Semantic Scholar search API."""
    def __init__(self,
        *,
        api: Optional[SemanticScholarApi] = None,
        num_results: int = 100,
        fields: List[str] = ['title', 'abstract'],
        verbose: bool = True
    ):
        """
        Args:
            api: The Semantic Scholar api service. Defaults to a new instance of :class:`~pyterrier_services.SemanticScholarApi`.
            num_results: The number of results to retrieve per query. Defaults to 100.
            fields: The fields to include in the retrieved results. Defaults to ['title', 'abstract'].
            verbose: Whether to log the progress. Defaults to True.
        """
        self.api = api or SemanticScholarApi()
        self.num_results = num_results
        self.fields = fields
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        return multi_query(
            paginated_search(
                http_error_retry(
                    partial(self.api.search, fields=self.fields)
                ),
                num_results=self.num_results,
            ),
            verbose=self.verbose,
            verbose_desc='SemanticScholarRetriever',
        )(inp)

    def fuse_rank_cutoff(self, k: int) -> Optional['SemanticScholarRetriever']:
        if k < self.num_results:
            return SemanticScholarRetriever(api=self.api, num_results=k, fields=self.fields, verbose=self.verbose)



import unittest
import pandas as pd
from pyterrier_services import SemanticScholarApi

class TestSemanticScholar(unittest.TestCase):
    def test_retriever(self):
        s2 = SemanticScholarApi()
        retr = s2.retriever(num_results=10)
        res = retr.search('PyTerrier')



import requests
import pyterrier as pt
import pandas as pd


def http_error_retry(fn, retries=3):
    def wrapped(*args, **kwargs):
        ex = None
        for i in range(retries):
            try:
                return fn(*args, **kwargs)
            except requests.exceptions.HTTPError as e:
                ex = e
        if ex is not None:
            raise ex
    return wrapped


def paginated_search(fn, num_results):
    def wrapped(query):
        pages = []
        count = 0
        offset = 0
        while count < num_results and offset is not None:
            page, offset = fn(query, offset=offset, limit=num_results-count, return_next=True)
            pages.append(page)
            count += len(page)
            if len(page) == 0:
                break
        return pd.concat(pages, ignore_index=True)
    return wrapped


def multi_query(fn, verbose=True, verbose_desc='retrieving'):
    def wrapped(inp):
        it = inp.itertuples(index=False)
        if verbose:
            it = pt.tqdm(it, desc=verbose_desc, unit='q', total=len(inp))
        res = []
        for query in it:
            query_res = fn(query.query)
            query_res = query_res.assign(**{k: v for k, v in query._asdict().items() if k not in query_res.columns})
            res.append(query_res)

        df = pd.concat(res, ignore_index=True)

        desired_order = ["qid", "query", "docno", "score", "rank"]

        # Add any remaining columns not in the desired order
        remaining_columns = [col for col in df.columns if col not in desired_order]
        new_order = desired_order + remaining_columns

        # Reorder the dataframe
        return df[new_order]
    return wrapped



import unittest
from pyterrier_services import DblpApi

class TestDblp(unittest.TestCase):
    def test_retriever(self):
        dblp = DblpApi()
        retr = dblp.retriever(num_results=10)
        retr.search('PyTerrier')

    def test_bibtex_loader(self):
        dblp = DblpApi()
        loader = dblp.bibtex_loader()
        loader([{'docno': 'conf/cikm/MacdonaldTMO21'}])
